package jalview.datamodel;

import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertFalse;
import static org.testng.AssertJUnit.assertTrue;

import jalview.io.AppletFormatAdapter;
import jalview.io.FormatAdapter;
import jalview.util.MapList;

import java.io.IOException;
import java.util.Iterator;

import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

/**
 * Unit tests for Alignment datamodel.
 * 
 * @author gmcarstairs
 *
 */
public class AlignmentTest
{
  // @formatter:off
  private static final String TEST_DATA = 
          "# STOCKHOLM 1.0\n" +
          "#=GS D.melanogaster.1 AC AY119185.1/838-902\n" +
          "#=GS D.melanogaster.2 AC AC092237.1/57223-57161\n" +
          "#=GS D.melanogaster.3 AC AY060611.1/560-627\n" +
          "D.melanogaster.1          G.AGCC.CU...AUGAUCGA\n" +
          "#=GR D.melanogaster.1 SS  ................((((\n" +
          "D.melanogaster.2          C.AUUCAACU.UAUGAGGAU\n" +
          "#=GR D.melanogaster.2 SS  ................((((\n" +
          "D.melanogaster.3          G.UGGCGCU..UAUGACGCA\n" +
          "#=GR D.melanogaster.3 SS  (.(((...(....(((((((\n" +
          "//";

  private static final String AA_SEQS_1 = 
          ">Seq1Name\n" +
          "K-QY--L\n" +
          ">Seq2Name\n" +
          "-R-FP-W-\n";

  private static final String CDNA_SEQS_1 = 
          ">Seq1Name\n" +
          "AC-GG--CUC-CAA-CT\n" +
          ">Seq2Name\n" +
          "-CG-TTA--ACG---AAGT\n";

  private static final String CDNA_SEQS_2 = 
          ">Seq1Name\n" +
          "GCTCGUCGTACT\n" +
          ">Seq2Name\n" +
          "GGGTCAGGCAGT\n";
  // @formatter:on

  private AlignmentI al;

  /**
   * Helper method to load an alignment and ensure dataset sequences are set up.
   * 
   * @param data
   * @param format
   *          TODO
   * @return
   * @throws IOException
   */
  protected AlignmentI loadAlignment(final String data, String format)
          throws IOException
  {
    Alignment a = new FormatAdapter().readFile(data,
            AppletFormatAdapter.PASTE, format);
    a.setDataset(null);
    return a;
  }

  /*
   * Read in Stockholm format test data including secondary structure
   * annotations.
   */
  @BeforeMethod
  public void setUp() throws IOException
  {
    al = loadAlignment(TEST_DATA, "STH");
    int i = 0;
    for (AlignmentAnnotation ann : al.getAlignmentAnnotation())
    {
      ann.setCalcId("CalcIdFor" + al.getSequenceAt(i).getName());
      i++;
    }
  }

  /**
   * Test method that returns annotations that match on calcId.
   */
  @Test
  public void testFindAnnotation_byCalcId()
  {
    Iterable<AlignmentAnnotation> anns = al
            .findAnnotation("CalcIdForD.melanogaster.2");
    Iterator<AlignmentAnnotation> iter = anns.iterator();
    assertTrue(iter.hasNext());
    AlignmentAnnotation ann = iter.next();
    assertEquals("D.melanogaster.2", ann.sequenceRef.getName());
    assertFalse(iter.hasNext());
  }

  @Test
  public void testDeleteAllAnnotations_includingAutocalculated()
  {
    AlignmentAnnotation aa = new AlignmentAnnotation("Consensus",
            "Consensus", 0.5);
    aa.autoCalculated = true;
    al.addAnnotation(aa);
    AlignmentAnnotation[] anns = al.getAlignmentAnnotation();
    assertEquals("Wrong number of annotations before deleting", 4,
            anns.length);
    al.deleteAllAnnotations(true);
    assertEquals("Not all deleted", 0, al.getAlignmentAnnotation().length);
  }

  @Test
  public void testDeleteAllAnnotations_excludingAutocalculated()
  {
    AlignmentAnnotation aa = new AlignmentAnnotation("Consensus",
            "Consensus", 0.5);
    aa.autoCalculated = true;
    al.addAnnotation(aa);
    AlignmentAnnotation[] anns = al.getAlignmentAnnotation();
    assertEquals("Wrong number of annotations before deleting", 4,
            anns.length);
    al.deleteAllAnnotations(false);
    assertEquals("Not just one annotation left", 1,
            al.getAlignmentAnnotation().length);
  }

  /**
   * Tests for realigning as per a supplied alignment: Dna as Dna.
   * 
   * Note: AlignedCodonFrame's state variables are named for protein-to-cDNA
   * mapping, but can be exploited for a general 'sequence-to-sequence' mapping
   * as here.
   * 
   * @throws IOException
   */
  @Test
  public void testAlignAs_dnaAsDna() throws IOException
  {
    // aligned cDNA:
    AlignmentI al1 = loadAlignment(CDNA_SEQS_1, "FASTA");
    // unaligned cDNA:
    AlignmentI al2 = loadAlignment(CDNA_SEQS_2, "FASTA");

    /*
     * Make mappings between sequences. The 'aligned cDNA' is playing the role
     * of what would normally be protein here.
     */
    AlignedCodonFrame acf = new AlignedCodonFrame();
    MapList ml = new MapList(new int[]
    { 1, 12 }, new int[]
    { 1, 12 }, 1, 1);
    acf.addMap(al2.getSequenceAt(0), al1.getSequenceAt(0), ml);
    acf.addMap(al2.getSequenceAt(1), al1.getSequenceAt(1), ml);
    al1.addCodonFrame(acf);

    ((Alignment) al2).alignAs(al1, false, true);
    assertEquals("GC-TC--GUC-GTA-CT", al2.getSequenceAt(0)
            .getSequenceAsString());
    assertEquals("-GG-GTC--AGG---CAGT", al2.getSequenceAt(1)
            .getSequenceAsString());
  }

  /**
   * Aligning protein from cDNA.
   * 
   * @throws IOException
   */
  @Test
  public void testAlignAs_proteinAsCdna() throws IOException
  {
    // see also AlignmentUtilsTests
    AlignmentI al1 = loadAlignment(CDNA_SEQS_1, "FASTA");
    AlignmentI al2 = loadAlignment(AA_SEQS_1, "FASTA");
    AlignedCodonFrame acf = new AlignedCodonFrame();
    MapList ml = new MapList(new int[]
    { 1, 12 }, new int[]
    { 1, 4 }, 3, 1);
    acf.addMap(al1.getSequenceAt(0), al2.getSequenceAt(0), ml);
    acf.addMap(al1.getSequenceAt(1), al2.getSequenceAt(1), ml);
    al2.addCodonFrame(acf);

    ((Alignment) al2).alignAs(al1, false, true);
    assertEquals("K-Q-Y-L-", al2.getSequenceAt(0).getSequenceAsString());
    assertEquals("-R-F-P-W", al2.getSequenceAt(1).getSequenceAsString());
  }

  /**
   * Test aligning cdna as per protein alignment.
   * 
   * @throws IOException
   */
  @Test
  public void testAlignAs_cdnaAsProtein() throws IOException
  {
    /*
     * Load alignments and add mappings for cDNA to protein
     */
    AlignmentI al1 = loadAlignment(CDNA_SEQS_1, "FASTA");
    AlignmentI al2 = loadAlignment(AA_SEQS_1, "FASTA");
    AlignedCodonFrame acf = new AlignedCodonFrame();
    MapList ml = new MapList(new int[]
    { 1, 12 }, new int[]
    { 1, 4 }, 3, 1);
    acf.addMap(al1.getSequenceAt(0), al2.getSequenceAt(0), ml);
    acf.addMap(al1.getSequenceAt(1), al2.getSequenceAt(1), ml);
    al2.addCodonFrame(acf);

    /*
     * Realign DNA; currently keeping existing gaps in introns only
     */
    ((Alignment) al1).alignAs(al2, false, true);
    assertEquals("ACG---GCUCCA------ACT", al1.getSequenceAt(0)
            .getSequenceAsString());
    assertEquals("---CGT---TAACGA---AGT", al1.getSequenceAt(1)
            .getSequenceAsString());
  }

  /**
   * Test aligning dna as per protein alignment, for the case where there are
   * introns (i.e. some dna sites have no mapping from a peptide).
   * 
   * @throws IOException
   */
  @Test
  public void testAlignAs_dnaAsProtein_withIntrons() throws IOException
  {
    /*
     * Load alignments and add mappings for cDNA to protein
     */
    String dna1 = "A-Aa-gG-GCC-cT-TT";
    String dna2 = "c--CCGgg-TT--T-AA-A";
    AlignmentI al1 = loadAlignment(">Seq1\n" + dna1 + "\n>Seq2\n" + dna2
            + "\n", "FASTA");
    AlignmentI al2 = loadAlignment(">Seq1\n-P--YK\n>Seq2\nG-T--F\n",
            "FASTA");
    AlignedCodonFrame acf = new AlignedCodonFrame();
    // Seq1 has intron at dna positions 3,4,9 so splice is AAG GCC TTT
    // Seq2 has intron at dna positions 1,5,6 so splice is CCG TTT AAA
    MapList ml1 = new MapList(new int[]
    { 1, 2, 5, 8, 10, 12 }, new int[]
    { 1, 3 }, 3, 1);
    acf.addMap(al1.getSequenceAt(0), al2.getSequenceAt(0), ml1);
    MapList ml2 = new MapList(new int[]
    { 2, 4, 7, 12 }, new int[]
    { 1, 3 }, 3, 1);
    acf.addMap(al1.getSequenceAt(1), al2.getSequenceAt(1), ml2);
    al2.addCodonFrame(acf);

    /*
     * Align ignoring gaps in dna introns and exons
     */
    ((Alignment) al1).alignAs(al2, false, false);
    assertEquals("---AAagG------GCCcTTT", al1.getSequenceAt(0)
            .getSequenceAsString());
    // note 1 gap in protein corresponds to 'gg-' in DNA (3 positions)
    assertEquals("cCCGgg-TTT------AAA", al1.getSequenceAt(1)
            .getSequenceAsString());

    /*
     * Reset and realign, preserving gaps in dna introns and exons
     */
    al1.getSequenceAt(0).setSequence(dna1);
    al1.getSequenceAt(1).setSequence(dna2);
    ((Alignment) al1).alignAs(al2, true, true);
    // String dna1 = "A-Aa-gG-GCC-cT-TT";
    // String dna2 = "c--CCGgg-TT--T-AA-A";
    // assumption: we include 'the greater of' protein/dna gap lengths, not both
    assertEquals("---A-Aa-gG------GCC-cT-TT", al1.getSequenceAt(0)
            .getSequenceAsString());
    assertEquals("c--CCGgg-TT--T------AA-A", al1.getSequenceAt(1)
            .getSequenceAsString());
  }
}
